import javax.swing.*;
import java.applet.Applet;
import java.awt.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;

// - Enum declarations -------------------------------------------------------
enum AD {
	INVALID, ANALOG, DIGITAL
}

enum MODE {
	INVALID, INPUT, OUTPUT
}

enum BITORDER {
	INVALID, MSBFIRST, LSBFIRST
}

// Needs to extend Applet for displaying it on a higher level...
public abstract class Core extends Applet implements Runnable, PinChangeListener {
	// - Constants, usabel for Testbed ---------------------------------------
	public static final MODE INPUT = MODE.INPUT;
	public static final MODE OUTPUT = MODE.OUTPUT;

	public static final int LEVEL_INVALID = 0;
	public static final int LOW = 1;
	public static final int HIGH = 2;

	public static final ASCIIMODE DEC = ASCIIMODE.DEC;
	public static final ASCIIMODE HEX = ASCIIMODE.HEX;
	public static final ASCIIMODE OCT = ASCIIMODE.OCT;
	public static final ASCIIMODE BIN = ASCIIMODE.BIN;
	public static final ASCIIMODE BYTE = ASCIIMODE.BYTE;

	public static final BITORDER MSBFIRST = BITORDER.MSBFIRST;
	public static final BITORDER LSBFIRST = BITORDER.LSBFIRST;

	// - Variables -----------------------------------------------------------
	private int[] pinValueAnalog;
	private int[] pinValueDigital;
	private MODE[] pinMode;

	private final ArrayList<PinChangeListener> changeListeners = new ArrayList<PinChangeListener>();
	private final HashMap<Integer, Pin> digitalIn = new HashMap<Integer, Pin>();
	private final HashMap<Integer, Pin> digitalOut = new HashMap<Integer, Pin>();
	private final HashMap<Integer, Pin> analogIn = new HashMap<Integer, Pin>();
	private final HashMap<Integer, Pin> analogOut = new HashMap<Integer, Pin>();
	private final long startMillis = System.currentTimeMillis();

	private Random random;
	private int maxPin = 0;

	int[] DIGITAL_OUT, DIGITAL_IN, PWM_OUT, ANALOG_IN;

	// - Semi-abstract functions ---------------------------------------------
	public abstract void setupBoard ();

	abstract void setup ();

	abstract void loop ();

	// - Internals -----------------------------------------------------------

	@Override
	public void update (Graphics g) {
		paint(g); // No DoubleBuffering, just paint over.
	}

	@Override
	public void init () {
		setBackground(Color.WHITE);

		setupBoard();

		pinValueAnalog = new int[maxPin + 1];
		pinValueDigital = new int[maxPin + 1];
		pinMode = new MODE[maxPin + 1];
		for (int t = 0; t <= maxPin; t++)
			pinValueAnalog[t] = 512;
		for (int t = 0; t <= maxPin; t++)
			pinValueDigital[t] = LEVEL_INVALID;
		for (int t = 0; t <= maxPin; t++)
			pinMode[t] = MODE.INVALID;

		new Thread(this).start(); // Disconnect ArduinoTestbed#loop() from frontend...
	}

	public void run () {
		setup(); // -> ArduinoTestbed...

		//noinspection InfiniteLoopStatement
		while (true) {
			loop(); // -> ArduinoTestbed...
			try {
				Thread.sleep(10);
			} catch (InterruptedException ignored) {
			}
			repaint();
		}
	}

	public Container add (final Container container, final PinChangeListener pinChangeListener) {
		return add(container, pinChangeListener, null);
	}

	public Container add (final Container container, final PinChangeListener comp, final Object constraints) {
		comp.addChangeListener(this);
		addChangeListener(comp);
		if (comp instanceof Pin) {
			final Pin pin = (Pin) comp;
			final int id = pin.getId();
			if (maxPin < id) // For generating an array, later...
				maxPin = id;
			if ((pin.getFeatures() & Pin.DIGITAL_IN) > 0)
				digitalIn.put(id, pin);
			if ((pin.getFeatures() & Pin.DIGITAL_OUT) > 0)
				digitalOut.put(id, pin);
			if ((pin.getFeatures() & Pin.ANALOG_IN) > 0)
				analogIn.put(id, pin);
			if ((pin.getFeatures() & Pin.ANALOG_OUT) > 0)
				analogOut.put(id, pin);
		} else {
			error(comp + " is not a Pin!");
		}
		if (constraints != null) {
			container.add((Component) comp, constraints); // Add to frontend
		} else {
			container.add((Component) comp); // Add to frontend
		}
		return (Container) comp;
	}

	public void setLevel (final Container container, final int level) {
		pinValueDigital[((Pin) container).getId()] = level;
	}

	public void setValue (final Container container, final int value) {
		pinValueAnalog[((Pin) container).getId()] = value;
	}

	public void addChangeListener (final PinChangeListener pinChangeListener) {
		changeListeners.add(pinChangeListener);
	}

	public static void error (final String string) {
		System.err.println(
			(System.currentTimeMillis() % 1000) + " " +
				Thread.currentThread().getStackTrace()[2].getClassName() + "(" +
				Thread.currentThread().getStackTrace()[2].getLineNumber() + "): " +
				string
		);
	}

	// - Wrapper Functions, Software -----------------------------------------
	public long millis () {
		return System.currentTimeMillis() - startMillis;
	}

	public long micros () {
		return (System.currentTimeMillis() - startMillis) / 1000;
	}

	public void delay (final int ms) {
		try {
			Thread.sleep(ms);
		}
		catch (InterruptedException ignored) {
		}
	}

	public void delayMicroseconds (final int us) {
		try {
			int ms = us / 1000;
			int ns = us % 1000 * 1000;
			Thread.sleep(ms, us);
		}
		catch (InterruptedException ignored) {
		}
	}

	public double min (final double x, final double y) {
		return Math.min(x, y);
	}

	public double max (final double x, final double y) {
		return Math.max(x, y);
	}

	public double abs (final double x) {
		return Math.abs(x);
	}

	public int constrain (final double x, final double a, final double b) {
		return (int) (Math.min(Math.max(x, a), b));
	}

	public int map (final int x, final int in_min, final int in_max, final int out_min, final int out_max) {
		return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
	}

	public double pow (final double base, final double exponent) {
		return Math.pow(base, exponent);
	}

	public double sq (final double x) {
		return Math.sqrt(x);
	}

	public double sqrt (final double x) {
		return Math.sqrt(x);
	}

	public double sin (final double rad) {
		return Math.sin(rad);
	}

	public double cos (final double rad) {
		return Math.cos(rad);
	}

	public double tan (final double rad) {
		return Math.tan(rad);
	}

	public void randomSeed (final long seed) {
		random = new Random(seed);
	}

	public long random (final long max) {
		return random(0, max);
	}

	public long random (final long min, final long max) {
		return random.nextInt((int) (max - min)) + min;
	}

	// - Wrapper Functions, Hardware -----------------------------------------
	public void pinMode (final int pin, final MODE mode) {
		pinMode[pin] = mode;
		switch (mode) {
			case INPUT:
				digitalIn.get(pin).setMode(mode);
				break;
			case OUTPUT:
				digitalOut.get(pin).setMode(mode);
				break;
			case INVALID:
				error("Pinmode for pin " + pin + " is INVALID!");
		}
	}

	public int digitalRead (final int pin) {
		if (pinMode[pin] != MODE.INPUT)
			error("Pinmode for pin " + pin + " is not INPUT!");
		if (pinValueDigital[pin] == LEVEL_INVALID)
			error("Pinlevel for pin " + pin + " is INVALID!");
		return pinValueDigital[pin];
	}

	public void digitalWrite (final int pin, final int value) {
		if (pinMode[pin] != MODE.OUTPUT)
			error("Pinmode for pin " + pin + " is not OUTPUT!");
		pinValueDigital[pin] = value;
		pinValueAnalog[pin] = 0;
		digitalOut.get(pin).setLevel(this, value);
	}

	// Don't check for pinmode, because analog pins are always in 'read'-mode...
	public int analogRead (final int pin) {
		analogIn.get(pin).hasBeenReadFrom(); // For changing the color!
		return pinValueAnalog[pin];
	}

	// Don't check for pinmode, because analog pins semms to always be in 'write'-mode...
	public void analogWrite (final int pin, final int value) {
		pinValueAnalog[pin] = value;
		pinValueDigital[pin] = LEVEL_INVALID;
		digitalOut.get(pin).setValue(this, value);
	}

	public void shiftOut (final int dataPin, final int clockPin, final BITORDER bitOrder, final int value) {
		// TODO
	}

	public long pulseIn (final int pin, final int value) {
		return 0; // TODO
	}

	public long pulseIn (final int pin, final int value, final int timeout) {
		return 0; // TODO
	}
}
